﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Joseph Helm and Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PackingEncoding;

namespace HeuristicLab.Problems.BinPacking {

  [Item("Packing-Solution Decoder", "An operator to create a concrete bin packing solution from a given encoding-instance.")]
  [StorableClass]
  public abstract class PackingSolutionDecoder<D, B, I> : SingleSuccessorOperator, IPackingSolutionDecoder
    where D : class, IPackingPosition
    where B : PackingShape<D>
    where I : PackingShape<D>, IPackingItem {

    public ILookupParameter<IRandom> RandomParameter {
      get { return (LookupParameter<IRandom>)Parameters["Random"]; }
    }
    private ILookupParameter<ItemList<I>> PackingItemMeasuresParameter {
      get { return (LookupParameter<ItemList<I>>)Parameters["PackingItemMeasures"]; }
    }
    private ILookupParameter<B> PackingBinMeasuresParameter {
      get { return (LookupParameter<B>)Parameters["PackingBinMeasures"]; }
    }
    public ILookupParameter<IItem> EncodedSolutionParameter {
      get { return (ILookupParameter<IItem>)Parameters["EncodedSolution"]; }
    }
    public ILookupParameter<PackingPlan<D,B,I>> PackingPlanParameter {
      get { return (ILookupParameter<PackingPlan<D, B, I>>)Parameters["PackingPlan"]; }
    }
    public ValueParameter<BoolValue> StackingConstraintsParameter {
      get { return (ValueParameter<BoolValue>)Parameters["StackingConstraint"]; }
    }


    public PackingSolutionDecoder() : base() {
      Parameters.Add(new LookupParameter<IItem>("EncodedSolution", "The new bin-packing solution represented as encoding."));
      Parameters.Add(new LookupParameter<PackingPlan<D, B, I>>("PackingPlan", "The decoded bin-packing solution represented as generalized packing plan."));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator which should be used for stochastic manipulation operators."));
      Parameters.Add(new LookupParameter<ItemList<I>>("PackingItemMeasures", "Packing-item data taken from the bin-packing problem-instance."));
      Parameters.Add(new LookupParameter<B>("PackingBinMeasures", "Packing-bin data taken from the bin-packing problem-instance."));
      Parameters.Add(new ValueParameter<BoolValue>("StackingConstraint", "A flag determining whether stacking constraints should be enforced or not.", new BoolValue(true)));
    }

    [StorableConstructor]
    protected PackingSolutionDecoder(bool deserializing) : base(deserializing) { }
    protected PackingSolutionDecoder(PackingSolutionDecoder<D,B,I> original, Cloner cloner)
      : base(original, cloner) {
    }

    public abstract PackingPlan<D, B, I> CreatePackingPlanFromEncoding(IItem solution, B binMeasures, ItemList<I> itemMeasures);

    public override IOperation Apply() {
      PackingPlan<D, B, I> result = CreatePackingPlanFromEncoding(EncodedSolutionParameter.ActualValue, PackingBinMeasuresParameter.ActualValue, PackingItemMeasuresParameter.ActualValue);
      PackingPlanParameter.ActualValue = result;
      return base.Apply();
    }
  }
}
